import {
    AbstractFactory,
    Container,
    DependencyRequirement,
    PropertyDependency,
    isInstantiable,
    DEPENDENCY_KEYS_METADATA_KEY,
    DEPENDENCY_KEYS_PROPERTY_METADATA_KEY, Instantiable, FactoryProducer,
} from '../../../di'
import {Collection, ErrorWithContext} from '../../../util'
import {Config} from '../../../service/Config'
import {ScopeRepository} from '../types'
import {ConfigScopeRepository} from './ConfigScopeRepository'

/**
 * A dependency injection factory that matches the abstract ScopeRepository class
 * and produces an instance of the configured repository driver implementation.
 */
@FactoryProducer()
export class ScopeRepositoryFactory extends AbstractFactory<ScopeRepository> {
    protected get config(): Config {
        return Container.getContainer().make<Config>(Config)
    }

    produce(): ScopeRepository {
        return new (this.getScopeRepositoryClass())()
    }

    match(something: unknown): boolean {
        return something === ScopeRepository
    }

    getDependencyKeys(): Collection<DependencyRequirement> {
        const meta = Reflect.getMetadata(DEPENDENCY_KEYS_METADATA_KEY, this.getScopeRepositoryClass())
        if ( meta ) {
            return meta
        }
        return new Collection<DependencyRequirement>()
    }

    getInjectedProperties(): Collection<PropertyDependency> {
        const meta = new Collection<PropertyDependency>()
        let currentToken = this.getScopeRepositoryClass()

        do {
            const loadedMeta = Reflect.getMetadata(DEPENDENCY_KEYS_PROPERTY_METADATA_KEY, currentToken)
            if ( loadedMeta ) {
                meta.concat(loadedMeta)
            }
            currentToken = Object.getPrototypeOf(currentToken)
        } while (Object.getPrototypeOf(currentToken) !== Function.prototype && Object.getPrototypeOf(currentToken) !== Object.prototype)

        return meta
    }

    /**
     * Return the instantiable class of the configured scope repository backend.
     * @protected
     * @return Instantiable<ScopeRepository>
     */
    protected getScopeRepositoryClass(): Instantiable<ScopeRepository> {
        const ScopeRepositoryClass = this.config.get('oauth2.repository.scope', ConfigScopeRepository)

        if ( !isInstantiable(ScopeRepositoryClass) || !(ScopeRepositoryClass.prototype instanceof ScopeRepository) ) {
            const e = new ErrorWithContext('Provided client repository class does not extend from @extollo/lib.ScopeRepository')
            e.context = {
                configKey: 'oauth2.repository.client',
                class: ScopeRepositoryClass.toString(),
            }
        }

        return ScopeRepositoryClass
    }
}
